@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 2001, 2002, 2003 Free Software Foundation, Inc.
+ * Copyright 2001, 2002, 2003, 2005, 2006 Free Software Foundation, Inc.
  * Written by Paolo Bonzini.
  *
  * This file is part of GNU Smalltalk.
@@ -17,6 +17,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -24,7 +46,7 @@
  * 
  * You should have received a copy of the GNU General Public License along with
  * GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
  *
  ***********************************************************************/
 
@@ -35,8 +57,14 @@
 # include <readline/history.h>
 #endif
 
-#define EMACS_PROCESS_MARKER	'\001'	/* ^A as marker -- random
-					   choice */
+typedef struct gst_file_segment
+{
+  OBJ_HEADER;
+  OOP fileName;
+  OOP startPos;
+  OOP length;
+}
+ *gst_file_segment;
 
 typedef struct string_stream
 {
@@ -58,7 +86,7 @@ typedef struct oop_stream
 {
   OOP  oop;
   char *buf;
-  const char *ptr;
+  char *ptr;
   const char *end;
 }
 oop_stream;
@@ -66,14 +94,20 @@ oop_stream;
 typedef struct input_stream
 {
   stream_type type;
-  char pushedBackChars[3];	/* the 3 buffered characters */
-  int pushedBackCount;		/* number of chars pushed back */
+  /* the 3 buffered characters */
+  char pushedBackChars[3];
+
+  /* number of chars pushed back */
+  int pushedBackCount;
+
   int line;
   int column;
-  const char *fileName;
-  int fileOffset;
+  const char *prompt;
+
   OOP fileNameOOP;		/* the full path name for file */
-  mst_Boolean prompt;
+  const char *fileName;
+  off_t fileOffset;
+
   union
   {
     unix_file_stream u_st_file;
@@ -95,9 +129,6 @@ static int my_getc (input_stream stream);
 /* Print a line indicator in front of an error message.  */
 static void line_stamp (int line);
 
-/* Close the given stream */
-static void my_close (input_stream stream);
-
 /* Allocate and push a new stream of type TYPE on the stack; the new
    stream is then available through IN_STREAM.  */
 static input_stream push_new_stream (stream_type type);
@@ -113,10 +144,8 @@ static int poll_and_read (int fd, char *buf, int n);
 static int change_str = -1;
 
 
-/* If true, the normal execution information is supressed, and the
-   prompt is emitted with a special marker character ahead of it to
-   let the process filter know that the execution has completed.  */
-mst_Boolean _gst_emacs_process = false;
+/* If true, readline is disabled.  */
+mst_Boolean _gst_no_tty = false;
 
 /* >= 1 if completions are enabled, < 1 if they are not.  Available
    for completeness even if Readline is not used.  */
@@ -181,16 +210,36 @@ static char **readline_match_symbols (char *text,
 void
 _gst_pop_stream (mst_Boolean closeIt)
 {
-  input_stream oldStream;
+  input_stream stream;
 
-  oldStream = in_stream;
+  stream = in_stream;
   in_stream = in_stream->prevStream;
 
-  if (closeIt && oldStream)
+  if (!IS_NIL (stream->fileNameOOP))
+    _gst_unregister_oop (stream->fileNameOOP);
+
+  switch (stream->type)
     {
-      my_close (oldStream);
-      xfree (oldStream);
+    case STREAM_STRING:
+      xfree (stream->st_str.strBase);
+      break;
+
+#ifdef HAVE_READLINE
+    case STREAM_READLINE:
+#endif /* HAVE_READLINE */
+    case STREAM_OOP:
+      xfree (stream->st_oop.buf);
+      _gst_unregister_oop (stream->st_oop.oop);
+      break;
+
+    case STREAM_FILE:
+      xfree (stream->st_file.buf);
+      if (closeIt)
+        close (stream->st_file.fd);
+      break;
     }
+
+  xfree (stream);
 }
 
 void
@@ -205,7 +254,7 @@ _gst_push_unix_file (int fd,
   newStream->st_file.ptr = newStream->st_file.buf;
   newStream->st_file.end = newStream->st_file.buf;
   newStream->fileName = fileName;
-  newStream->prompt = isatty (fd);
+  newStream->fileOffset = lseek (fd, 0, SEEK_CUR);
 }
 
 void
@@ -219,7 +268,6 @@ _gst_push_stream_oop (OOP oop)
   newStream->st_oop.ptr = NULL;
   newStream->st_oop.end = NULL;
   newStream->fileName = "a Smalltalk Stream";
-  newStream->prompt = false;
 
   _gst_register_oop (oop);
 }
@@ -234,7 +282,6 @@ _gst_push_smalltalk_string (OOP stringOOP)
   newStream->st_str.strBase = (char *) _gst_to_cstring (stringOOP);
   newStream->st_str.str = newStream->st_str.strBase;
   newStream->fileName = "a Smalltalk string";
-  newStream->prompt = false;
 }
 
 void
@@ -247,7 +294,6 @@ _gst_push_cstring (const char *string)
   newStream->st_str.strBase = xstrdup (string);
   newStream->st_str.str = newStream->st_str.strBase;
   newStream->fileName = "a C string";
-  newStream->prompt = false;
 }
 
 void
@@ -256,7 +302,7 @@ _gst_push_stdin_string (void)
 #ifdef HAVE_READLINE
   input_stream newStream;
 
-  if (_gst_emacs_process || !isatty (0))
+  if (_gst_no_tty)
     {
 #endif
       _gst_push_unix_file (0, "stdin");
@@ -264,12 +310,15 @@ _gst_push_stdin_string (void)
 #ifdef HAVE_READLINE
     }
 
-  newStream = push_new_stream (STREAM_READLINE);
+  if (count == 0)
+    _gst_add_all_symbol_completions ();
 
-  newStream->st_str.strBase = NULL;	/* force readline() but no free() */
-  newStream->st_str.str = NULL;
+  newStream = push_new_stream (STREAM_READLINE);
+  newStream->fileOffset = 0;
+  newStream->st_oop.buf = NULL;
+  newStream->st_oop.ptr = NULL;
+  newStream->st_oop.end = NULL;
   newStream->fileName = "stdin";	/* that's where we get input from */
-  newStream->prompt = true;
 #endif
 }
 
@@ -283,28 +332,55 @@ push_new_stream (stream_type type)
   newStream->pushedBackCount = 0;
   newStream->line = 1;
   newStream->column = 0;
-  newStream->fileOffset = 0;
+  newStream->fileOffset = -1;
   newStream->type = type;
+  newStream->fileName = NULL;
+  newStream->prompt = NULL;
   newStream->fileNameOOP = _gst_nil_oop;
   newStream->prevStream = in_stream;
   in_stream = newStream;
-
   return (newStream);
 }
 
 
 void
 _gst_set_stream_info (int line,
-		      const char *fileName,
+		      OOP fileNameOOP,
 		      int fileOffset)
 {
   in_stream->line = line;
   in_stream->column = 0;
-  if (in_stream->type == STREAM_FILE)
+  if (!IS_NIL (fileNameOOP))
     {
-      in_stream->fileName = fileName;
-      in_stream->fileOffset = fileOffset;
+      in_stream->fileName = _gst_to_cstring (fileNameOOP);
+      _gst_register_oop (in_stream->fileNameOOP);
     }
+
+  in_stream->fileNameOOP = fileNameOOP;
+  in_stream->fileOffset = fileOffset;
+}
+ 
+void
+refill_stream (input_stream stream, char *buf, int new_line)
+{
+  size_t old_size = stream->st_oop.ptr - stream->st_oop.buf;
+  size_t size = old_size + strlen (buf);
+
+  /* Leave space for the '\0' at the end.  */
+  stream->st_oop.buf = xrealloc (stream->st_oop.buf, size + new_line + 1);
+  stream->st_oop.ptr = stream->st_oop.buf + old_size;
+  stream->st_oop.end = stream->st_oop.buf + size + new_line;
+
+  memcpy (stream->st_oop.ptr, buf, size - old_size);
+  if (new_line)
+    {
+      stream->st_oop.ptr[size - old_size] = '\n';
+      stream->st_oop.ptr[size - old_size + 1] = '\0';
+    }
+  else
+    stream->st_oop.ptr[size - old_size] = '\0';
+
+  free (buf);
 }
 
 int
@@ -327,30 +403,20 @@ my_getc (input_stream stream)
       /* Refill the buffer...  */
       if (stream->st_oop.ptr == stream->st_oop.end)
 	{
-	  if (stream->st_oop.buf)
-	    {
-	      xfree (stream->st_oop.buf);
-	      stream->st_oop.buf = NULL;
-	    }
+	  char *buf;
+	  _gst_msg_sendf(&buf, "%s %o nextHunk", stream->st_oop.oop);
+	  if (!buf || !*buf)
+	    return EOF;
 
-	  _gst_msg_sendf(stream->st_oop.buf, "%s %o nextHunk",
-			 stream->st_oop.oop);
-
-	  stream->st_oop.end = stream->st_oop.buf +
-	    strlen (stream->st_oop.buf);
-	  stream->st_oop.ptr = stream->st_oop.buf;
+	  refill_stream (stream, buf, false);
 	}
 
-      return (stream->st_oop.ptr == stream->st_oop.end)
-	? EOF : (unsigned char) *stream->st_oop.ptr++;
+      return (unsigned char) *stream->st_oop.ptr++;
 
     case STREAM_FILE:
       if (in_stream->column == 0 && in_stream->prompt)
 	{
-	  if (_gst_emacs_process)
-	    printf ("%c", EMACS_PROCESS_MARKER);
-
-	  printf ("st> ");
+	  printf ("%s", in_stream->prompt);
 	  fflush(stdout);
 	}
 
@@ -361,6 +427,7 @@ my_getc (input_stream stream)
 	  if (n < 0)
 	    n = 0;
 
+	  stream->fileOffset += stream->st_file.ptr - stream->st_file.buf;
 	  stream->st_file.end = stream->st_file.buf + n;
 	  stream->st_file.ptr = stream->st_file.buf;
 	}
@@ -370,55 +437,20 @@ my_getc (input_stream stream)
 
 #ifdef HAVE_READLINE
     case STREAM_READLINE:
-      {
-	char *r_line;
-	int r_len;
+      /* Refill the buffer...  */
+      if (stream->st_oop.ptr == stream->st_oop.end)
+	{
+	  char *buf = readline (in_stream->prompt
+				? (char *) in_stream->prompt
+				: (char *) "");
+	  if (!buf)
+	    return EOF;
 
-	if (stream->st_str.strBase)
-	  {
-	    ic = (unsigned char) *stream->st_str.str++;
-	    if (ic)
-	      return (ic);
+	  add_history (buf);
+	  refill_stream (stream, buf, true);
+	}
 
-	    /* If null, read a new line */
-	  }
-
-	if (stream->st_str.strBase)
-	  {
-	    xfree (stream->st_str.strBase);
-	    stream->st_str.strBase = NULL;
-	  }
-	r_line = readline ((char *) "st> ");
-	if (!r_line)
-	  {
-	    /* return value of NULL indicates EOF */
-	    return (EOF);
-	  }
-	if (*r_line)
-	  {
-	    /* add only non-empty lines.  */
-	    add_history (r_line);
-	  }
-
-	/* tack on the newline, not returned by readline() */
-	r_len = strlen (r_line);
-	r_line = xrealloc (r_line, (unsigned) (r_len + 2));
-	if (!r_line)
-	  {
-	    _gst_errorf ("Out of memory reallocating linebuffer space");
-	    stream->st_str.str = stream->st_str.strBase = NULL;
-	    ic = '\n';		/* return a newline ...  */
-	  }
-	else
-	  {
-	    r_line[r_len] = '\n';
-	    r_line[r_len + 1] = '\0';
-	    stream->st_str.str = stream->st_str.strBase = r_line;
-	    ic = (unsigned char) *stream->st_str.str++;
-	  }
-
-	break;
-      }
+      return (unsigned char) *stream->st_oop.ptr++;
 #endif /* HAVE_READLINE */
 
     default:
@@ -428,49 +460,11 @@ my_getc (input_stream stream)
   return (ic);
 }
 
-void
-my_close (input_stream stream)
-{
-  if (!IS_NIL (stream->fileNameOOP))
-    _gst_unregister_oop (stream->fileNameOOP);
-
-  switch (stream->type)
-    {
-    case STREAM_STRING:
-      xfree (stream->st_str.strBase);
-      break;
-
-    case STREAM_OOP:
-      xfree (stream->st_oop.buf);
-      _gst_unregister_oop (stream->st_oop.oop);
-      break;
-
-    case STREAM_FILE:
-      xfree (stream->st_file.buf);
-      close (stream->st_file.fd);
-      break;
-
-#ifdef HAVE_READLINE
-    case STREAM_READLINE:
-      if (stream->st_str.strBase)
-	{
-	  xfree (stream->st_str.strBase);
-	  stream->st_str.strBase = NULL;
-	}
-      break;
-#endif /* HAVE_READLINE */
-
-    default:
-      _gst_errorf ("Bad stream type passed to my_close");
-      _gst_had_error = true;
-    }
-}
-
 
 mst_Boolean
 _gst_get_cur_stream_prompt (void)
 {
-  return (in_stream && in_stream->prompt);
+  return in_stream && in_stream->prompt;
 }
 
 stream_type
@@ -484,13 +478,73 @@ _gst_get_cur_stream_type (void)
 }
 
 OOP
-_gst_get_cur_string (void)
+_gst_get_source_string (off_t startPos, off_t endPos)
 {
-  if (in_stream && in_stream->type == STREAM_STRING)
-    return (_gst_string_new (in_stream->st_str.strBase));
+  char *p;
+  OOP result;
+  int size;
 
-  else
+  if (!in_stream)
     return (_gst_nil_oop);
+
+  if (startPos != -1 && !_gst_get_cur_stream_prompt ())
+    {
+      OOP fileName;
+      gst_file_segment fileSegment;
+      inc_ptr incPtr;
+
+      incPtr = INC_SAVE_POINTER ();
+      fileName = _gst_get_cur_file_name ();
+      INC_ADD_OOP (fileName);
+
+      fileSegment = (gst_file_segment) new_instance (_gst_file_segment_class,
+                                                     &result);
+
+      fileSegment->fileName = fileName;
+      fileSegment->startPos = from_c_int_64 (startPos);
+      fileSegment->length = from_c_int_64 (endPos - startPos);
+
+      assert (to_c_int_64 (fileSegment->length) >= 0);
+      INC_RESTORE_POINTER (incPtr);
+      return (result);
+    }
+
+  switch (in_stream->type)
+    {
+    case STREAM_STRING:
+      p = in_stream->st_str.strBase;
+      break;
+
+#ifdef HAVE_READLINE
+    case STREAM_READLINE:
+#endif /* HAVE_READLINE */
+    case STREAM_OOP:
+      p = in_stream->st_oop.buf;
+      break;
+
+    default:
+      return (_gst_nil_oop);
+    }
+
+   if (startPos == -1) 
+    result = _gst_string_new (p);
+  else
+    result = _gst_counted_string_new (p + (startPos - in_stream->fileOffset),
+				      endPos - startPos);
+
+  if (in_stream->type != STREAM_STRING)
+    {
+      /* Copy back to the beginning of the buffer to save memory.  */
+      size = in_stream->st_oop.end - in_stream->st_oop.ptr;
+      if (size)
+	memmove (in_stream->st_oop.buf, in_stream->st_oop.ptr, size);
+      in_stream->st_oop.buf[size] = 0;
+      in_stream->fileOffset += in_stream->st_oop.ptr - in_stream->st_oop.buf;
+      in_stream->st_oop.ptr = in_stream->st_oop.buf;
+      in_stream->st_oop.end = in_stream->st_oop.buf + size;
+    }
+
+  return result;
 }
 
 OOP
@@ -518,17 +572,6 @@ _gst_get_cur_file_name (void)
   return (in_stream->fileNameOOP);
 }
 
-#ifdef HAVE_READLINE
-OOP
-_gst_get_cur_readline (void)
-{
-  if (in_stream && in_stream->type == STREAM_READLINE)
-    return (_gst_string_new (in_stream->st_str.strBase));
-
-  else
-    return (_gst_nil_oop);
-}
-#endif /* HAVE_READLINE */
 
 
 void
@@ -631,11 +674,21 @@ _gst_yyerror (const char *s)
   _gst_errorf ("%s", s);
 }
 
-void
-_gst_get_location (int *x, int *y)
+YYLTYPE
+_gst_get_location (void)
 {
-  *x = in_stream->column;
-  *y = in_stream->line;
+  YYLTYPE loc;
+  loc.first_line = in_stream->line;
+  loc.first_column = in_stream->column;
+
+  if (!in_stream || in_stream->fileOffset == -1)
+    loc.file_offset = -1;
+  else
+    /* Subtract 1 to mark the position of the last character we read.  */
+    loc.file_offset = (in_stream->st_file.ptr - in_stream->st_file.buf
+		       + in_stream->fileOffset - 1);
+
+  return loc;
 }
 
 void
@@ -650,6 +703,8 @@ line_stamp (int line)
 	{
 	  if (in_stream->fileName)
 	    fprintf (stderr, "%s:", in_stream->fileName);
+	  else if (!IS_NIL (in_stream->fileNameOOP))
+	    fprintf (stderr, "%#O:", in_stream->fileNameOOP);
 
 	  fprintf (stderr, "%d: ", line);
 	}
@@ -678,33 +733,6 @@ line_stamp (int line)
 }
 
 
-int
-_gst_get_cur_file_pos (void)
-{
-  /* ### not sure about this particular use -- the current file pos
-     must be from the change log, but currently the FileSegments we
-     produce are from the input file... so they lose. The original
-     reason for implementing the change log was to make recompiles of 
-     a method lying in a changed file work... but this way not even
-     recompiles of methods lying in an unchanged file do work! */
-#ifdef not_working_for_now
-  if (change_str != -1)
-    return (lseek (change_str, 0, SEEK_CUR));
-  else
-#endif
-
-  if (in_stream && in_stream->type == STREAM_FILE)
-    {
-      return (lseek (in_stream->st_file.fd, 0, SEEK_CUR)
-	      + in_stream->st_file.ptr - in_stream->st_file.end
-	      + in_stream->fileOffset);
-    }
-  else
-    return (-1);
-}
-
-
-
 int
 _gst_next_char (void)
 {
@@ -772,7 +800,7 @@ poll_and_read (int fd, char *buf, int n)
       errno = 0;
       result = poll (&pfd, 1, -1); /* Infinite wait */
     }
-  while (result == 0
+  while ((result == 0 && (pfd.revents & POLLHUP) == 0)
 	 || ((result == -1) && (errno == EINTR)));
 
   if (pfd.revents & POLLIN)
@@ -789,6 +817,53 @@ poll_and_read (int fd, char *buf, int n)
     return -1;
 }
 
+void
+_gst_process_stdin (const char *prompt)
+{
+  if (_gst_verbosity == 3 || isatty (0))
+    {
+      printf ("GNU Smalltalk ready\n\n");
+      fflush (stdout);
+    }
+
+  _gst_non_interactive = false;
+  _gst_push_stdin_string ();
+  if (isatty (0))
+    in_stream->prompt = prompt;
+  _gst_parse_stream (false);
+  _gst_pop_stream (true);
+  _gst_non_interactive = true;
+}
+
+mst_Boolean
+_gst_process_file (const char *fileName, enum gst_file_dir dir)
+{
+  enum undeclared_strategy old;
+  int fd;
+  char *f;
+
+  f = _gst_find_file (fileName, dir);
+  if (!f)
+    return false;
+
+  fd = _gst_open_file (f, "r");
+  if (fd != -1)
+    {
+      if (_gst_verbosity == 3)
+	printf ("Processing %s\n", f);
+
+      old = _gst_set_undeclared (UNDECLARED_GLOBALS);
+      _gst_push_unix_file (fd, f);
+      _gst_parse_stream (false);
+      _gst_pop_stream (true);
+      _gst_set_undeclared (old);
+    }
+
+  xfree (f);
+  return (true);
+}
+
+
 #ifdef HAVE_READLINE
 /* Find apostrophes and double them */
 char *
@@ -1091,9 +1166,6 @@ _gst_initialize_readline (void)
 
   /* Set up to use read to read from stdin */
   rl_getc_function = readline_getc;
-
-  if (count == 0)
-    _gst_add_all_symbol_completions ();
 }
 
 #endif /* HAVE_READLINE */
